// SPDX-License-Identifier: MIT

use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use serde::Deserialize;

use crate::common;
use crate::common::{container::ContainerInfo, html_template::HtmlTemplate};
use crate::error::PageResult;

#[derive(Deserialize)]
struct ContainerParam {
    name: String,
}

#[derive(Template)]
#[template(path = "../src/container/templates/container.html")]
struct ContainerTemplate {
    container_info: ContainerInfo,
}

#[derive(Template)]
#[template(path = "../src/container/templates/container_logs.html")]
struct ContainerLogsTemplate {
    container_name: String,
    container_logs: String,
}

pub trait ContainerRoutes {
    fn container_routes(self) -> Self;
}

impl ContainerRoutes for Router {
    fn container_routes(self) -> Self {
        self.route("/container", get(container))
            .route("/container_start", post(container_start))
            .route("/container_stop", post(container_stop))
            .route("/container_log", get(container_log))
    }
}

async fn container(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let container_info = ContainerInfo::get().await?;
    let template = ContainerTemplate { container_info };
    Ok(HtmlTemplate::new("コンテナ管理", template).into_response())
}

async fn container_start(
    session: ReadableSession,
    Form(container_param): Form<ContainerParam>,
) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["podman.sh", "start", &container_param.name];
    common::exec_command(args).await?;

    Ok(Redirect::to("/container").into_response())
}

async fn container_stop(
    session: ReadableSession,
    Form(container_param): Form<ContainerParam>,
) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["podman.sh", "stop", &container_param.name];
    common::exec_command(args).await?;

    Ok(Redirect::to("/container").into_response())
}

async fn container_log(
    session: ReadableSession,
    Form(container_param): Form<ContainerParam>,
) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["podman.sh", "logs", &container_param.name];
    let output = common::exec_command(args).await?;

    let template = ContainerLogsTemplate {
        container_name: container_param.name,
        container_logs: String::from_utf8_lossy(&output.stdout).trim().to_string(),
    };
    Ok(HtmlTemplate::new("コンテナログ", template).into_response())
}
