// SPDX-License-Identifier: MIT

use anyhow::Result;
use axum::{http::StatusCode, Router};
use axum_sessions::{async_session::MemoryStore, SessionLayer};
use rand::Rng;
use std::time::Duration;
use tower_http::services::ServeDir;

mod args;
mod common;
mod connection;
mod container;
mod dhcp;
mod error;
mod ip_filter;
mod login;
mod nat;
mod nwstatbar;
mod server;
mod status;
mod swu;
mod top;
mod vpn;
mod wlan;
mod wwan;
use connection::ConnectionRoutes;
use container::ContainerRoutes;
use dhcp::DhcpRoutes;
use login::LoginRoutes;
use nat::NatRoutes;
use nwstatbar::NWStatBarRoutes;
use status::StatusRoutes;
use swu::SwuRoutes;
use top::TopRoutes;
use vpn::VpnRoutes;
use wlan::WlanRoutes;
use wwan::WwanRoutes;

#[cfg(debug_assertions)]
const STATIC_RESOURCE_DIR: &str = "static";

#[cfg(not(debug_assertions))]
const STATIC_RESOURCE_DIR: &str = "/usr/share/abos-web/static";

async fn fallback_404() -> (StatusCode, &'static str) {
    (StatusCode::NOT_FOUND, "Page not found")
}

#[tokio::main]
async fn main() -> Result<()> {
    let store = MemoryStore::new();
    let secret = rand::thread_rng().gen::<[u8; 128]>();
    let session_layer = SessionLayer::new(store, &secret)
        .with_secure(false)
        .with_session_ttl(Some(Duration::from_secs(60 * 60 * 12)));
    common::Config::load().await?;
    login::password_throttle_init();
    args::args_init();

    let app = Router::new()
        .connection_routes()
        .container_routes()
        .dhcp_routes()
        .login_routes()
        .nat_routes()
        .nwstatbar_routes()
        .status_routes()
        .swu_routes()
        .top_routes()
        .vpn_routes()
        .wlan_routes()
        .wwan_routes()
        .nest_service("/static", ServeDir::new(STATIC_RESOURCE_DIR))
        .layer(session_layer)
        .route_layer(axum::middleware::from_fn(ip_filter::ipaddr_filter))
        .fallback(fallback_404);

    server::run_server(app).await;
    Ok(())
}
