// SPDX-License-Identifier: MIT

use anyhow::{Context, Result};
use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use futures::future::join_all;
use ipnet::Ipv4Net;
use serde::{Deserialize, Serialize};
use std::{collections::HashMap, net::Ipv4Addr, str::FromStr};

use crate::common;
use crate::common::html_template::HtmlTemplate;
use crate::common::networkmanager::{nmcon_act, nmconmap, parse_nmcli_t_line, NMCon};
use crate::error::{ErrorStringResult, PageResult};

#[derive(Deserialize)]
struct WlanParam {
    essid: String,
    other_essid: String,
    password: String,
    security: String,
    method: String,
    ip_addr: String,
    netmask: String,
    gateway_addr: String,
    p_dns_addr: String,
    s_dns_addr: String,
}

#[derive(Deserialize, Debug)]
struct WlanSaved<'a> {
    uuid: &'a str,
    ssid: &'a str,
    autoconnect: &'a str,
}

#[derive(Deserialize, Serialize, Debug)]
struct WlanSSID {
    id: String,
    strength: u8,
}

#[derive(Deserialize, Debug)]
struct WlanUUID {
    con_uuid: String,
}

#[derive(Deserialize, Debug)]
struct APParam {
    bridge_addr: String,
    netmask: String,
    hw_mode: String,
    channel: String,
    ssid: String,
    wpa_passphrase: String,
}

#[derive(Deserialize, Debug)]
struct APDelParam {}

#[derive(Template)]
#[template(path = "../src/wlan/templates/wlan.html")]
struct WlanTemplate<'a> {
    ssid_list: Vec<WlanSSID>,
    wlan_saved: Vec<WlanSaved<'a>>,
    uuid: &'a str,
    ssid: &'a str,
    security: &'a str,
    current_ip_addr: String,
    current_gw_addr: String,
    current_netmask: String,
    interface: String,
    ap_info: HashMap<&'static str, String>,
    ap_br_interface: String,
}

#[derive(Template)]
#[template(path = "../src/wlan/templates/no_wlan.html")]
struct NoWlanTemplate {}

pub trait WlanRoutes {
    fn wlan_routes(self) -> Self;
}

impl WlanRoutes for Router {
    fn wlan_routes(self) -> Self {
        self.route("/wlan", get(wlan))
            .route("/wlan_scan", get(wlan_scan))
            .route("/wlan_setup", post(wlan_setup))
            .route("/wlan_save", post(wlan_save))
            .route("/wlan_delete", post(wlan_delete))
            .route("/wlan_connect", post(wlan_connect))
            .route("/wlan_disconnect", post(wlan_disconnect))
            .route("/ap_setup", post(ap_setup))
            .route("/ap_delete", post(ap_delete))
    }
}

fn wlan_level(strength: &str) -> Option<u8> {
    let signal_quality: u8 = strength.parse().ok()?;
    match signal_quality {
        0..=5 => Some(0),
        6..=30 => Some(1),
        31..=55 => Some(2),
        56..=79 => Some(3),
        80..=100 => Some(4),
        _ => None,
    }
}

async fn wlan(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let Some(interface) = common::wlan_interface() else {
        return Ok(HtmlTemplate::new("WLAN設定", NoWlanTemplate{}).into_response());
    };

    let current_addr = common::current_address(interface, interface).await?;
    let ap_info = common::ap_info().await?;

    let wlan: Vec<NMCon> = join_all(
        nmconmap()
            .await?
            .into_values()
            .filter(|c| c.ctype == "802-11-wireless")
            .map(|c| async move { NMCon::from_id(&c.uuid).await }),
    )
    .await
    .into_iter()
    .filter_map(|c| c.ok())
    .collect();

    let na = "--";
    let mut uuid = na;
    let mut ssid = na;
    let mut security = na;
    let mut wlan_saved = Vec::new();

    for con in wlan.iter() {
        if con.state == "activated" {
            uuid = con.prop("connection.uuid").unwrap_or(na);
            ssid = con.prop("802-11-wireless.ssid").unwrap_or(na);
            security = con.prop("802-11-wireless-security.key-mgmt").unwrap_or(na);
        }

        let data = WlanSaved {
            uuid: con.prop("connection.uuid").unwrap_or(na),
            ssid: con.prop("802-11-wireless.ssid").unwrap_or(na),
            autoconnect: con.prop("connection.autoconnect").unwrap_or(na),
        };
        wlan_saved.push(data);
    }
    wlan_saved.sort_unstable_by(|a, b| a.ssid.cmp(b.ssid));

    let output = common::exec_command(&["get_ssid_list.sh", "--no-rescan"]).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let ssid_list = stdout
        .lines()
        .map(parse_nmcli_t_line)
        .filter_map(|a| {
            Some(WlanSSID {
                id: a.get(0)?.to_string(),
                strength: wlan_level(a.get(1)?)?,
            })
        })
        .collect();

    let template = WlanTemplate {
        ssid_list,
        wlan_saved,
        uuid,
        ssid,
        security,
        current_ip_addr: current_addr["current_ip_addr"].to_owned(),
        current_netmask: current_addr["current_netmask"].to_owned(),
        current_gw_addr: current_addr["current_gw_addr"].to_owned(),
        interface: interface.to_string(),
        ap_info,
        ap_br_interface: common::AP_BRIDGE_INTERFACE.to_string(),
    };
    Ok(HtmlTemplate::new("WLAN設定", template).into_response())
}

async fn wlan_scan(session: ReadableSession) -> ErrorStringResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }
    let output = common::exec_command(&["get_ssid_list.sh", "--rescan"]).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let ssid_list: Vec<WlanSSID> = stdout
        .lines()
        .map(parse_nmcli_t_line)
        .filter_map(|a| {
            Some(WlanSSID {
                id: a.get(0)?.to_string(),
                strength: wlan_level(a.get(1)?)?,
            })
        })
        .collect();
    let list = serde_json::to_string(&ssid_list).context("Could not format json.")?;
    let mut res = list.into_response();
    res.headers_mut()
        .insert("content-type", "application/json".parse().unwrap());
    Ok(res)
}

fn wlan_save_args(args: &mut Vec<String>, wlan_params: WlanParam) -> Result<()> {
    let interface = common::wlan_interface().context("Could not find wlan network interface.")?;
    let mut essid = wlan_params.essid;
    if essid == "その他" {
        essid = wlan_params.other_essid;
    }
    args.push(essid);
    args.push(wlan_params.security);
    args.push(wlan_params.password);
    args.push(interface.to_string());
    args.push(wlan_params.method.clone());

    if wlan_params.method == "manual" {
        let netmask = if wlan_params.netmask.is_empty() {
            "255.255.255.255"
        } else {
            &wlan_params.netmask
        };
        let ip_addr = Ipv4Addr::from_str(&wlan_params.ip_addr)
            .with_context(|| format!("Could not parse addr: {}", wlan_params.ip_addr))?;
        let netmask = Ipv4Addr::from_str(netmask)
            .with_context(|| format!("Could not parse network: {}", netmask))?;
        let net = Ipv4Net::with_netmask(ip_addr, netmask)
            .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;

        args.push("-a".to_string());
        args.push(net.to_string());
        if !wlan_params.gateway_addr.is_empty() {
            args.push("-g".to_string());
            args.push(wlan_params.gateway_addr);
        }
        if !wlan_params.p_dns_addr.is_empty() {
            let mut dns = wlan_params.p_dns_addr;
            if !wlan_params.s_dns_addr.is_empty() {
                dns = format!("{},{}", dns, wlan_params.s_dns_addr);
            }
            args.push("-d".to_string());
            args.push(dns);
        }
    }
    Ok(())
}

async fn wlan_setup(session: ReadableSession, Form(wlan_param): Form<WlanParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let mut args = vec!["wlan_setup.sh".to_string()];
    wlan_save_args(&mut args, wlan_param)?;
    args.push("--connect".to_string());

    common::exec_command(&args).await?;
    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_save(session: ReadableSession, Form(wlan_param): Form<WlanParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let mut args = vec!["wlan_setup.sh".to_string()];
    wlan_save_args(&mut args, wlan_param)?;

    common::exec_command(&args).await?;
    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_delete(session: ReadableSession, Form(wlan): Form<WlanUUID>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    nmcon_act("delete", &wlan.con_uuid).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_connect(session: ReadableSession, Form(wlan): Form<WlanUUID>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    nmcon_act("up", &wlan.con_uuid).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_disconnect(session: ReadableSession, Form(wlan): Form<WlanUUID>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    nmcon_act("down", &wlan.con_uuid).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_setup(session: ReadableSession, Form(ap_param): Form<APParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let netmask = if ap_param.netmask.is_empty() {
        "255.255.255.255"
    } else {
        &ap_param.netmask
    };
    let ip_addr = Ipv4Addr::from_str(&ap_param.bridge_addr)
        .with_context(|| format!("Could not parse addr: {}", &ap_param.bridge_addr))?;
    let netmask = Ipv4Addr::from_str(netmask)
        .with_context(|| format!("Could not parse network: {}", netmask))?;
    let net = Ipv4Net::with_netmask(ip_addr, netmask)
        .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;
    let interface = common::ap_interface().context("Could not find wlan network interface.")?;

    let args = &[
        "ap_setup.sh",
        common::AP_BRIDGE_NAME,
        &net.to_string(),
        &ap_param.hw_mode,
        &ap_param.channel,
        &ap_param.ssid,
        &ap_param.wpa_passphrase,
        interface,
    ];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_delete(session: ReadableSession, Form(_): Form<APDelParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let interface = common::ap_interface().context("Could not find wlan network interface.")?;

    let args = &["ap_delete.sh", "abos_web_br_ap", interface];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}
